package main

import (
	"context"
	"log"
	"net/http"
	"time"

	"github.com/kelseyhightower/envconfig"
	"github.com/penkovski/graceful"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"

	"code.vereign.com/gaiax/tsa/infohub/internal/cache"
	"code.vereign.com/gaiax/tsa/infohub/internal/config"
	"code.vereign.com/gaiax/tsa/infohub/internal/httpapi"
	"code.vereign.com/gaiax/tsa/infohub/internal/tasks"
)

var Version = "0.0.0+development"

func main() {
	var cfg config.Config
	if err := envconfig.Process("", &cfg); err != nil {
		log.Fatal("error loading configuration:", err)
	}

	logger, err := createLogger()
	if err != nil {
		log.Fatalln(err)
	}
	defer logger.Sync() //nolint:errcheck

	logger.Info("starting information hub service", zap.String("version", Version))

	cache := cache.New(
		cfg.Cache.Host,
		cfg.Cache.Port,
		cfg.Cache.Pass,
		cfg.Cache.DB,
		logger,
	)

	tasks := tasks.New(cache, logger)
	go tasks.Start(context.Background())

	handler := httpapi.New(tasks, logger)
	httpsrv := &http.Server{
		Addr:         cfg.HTTP.Host + ":" + cfg.HTTP.Port,
		Handler:      handler,
		ReadTimeout:  cfg.HTTP.ReadTimeout,
		WriteTimeout: cfg.HTTP.WriteTimeout,
		IdleTimeout:  cfg.HTTP.IdleTimeout,
	}
	if err := graceful.Start(httpsrv, 20*time.Second); err != nil {
		logger.Fatal("server shutdown error", zap.Error(err))
	}

}

func createLogger(opts ...zap.Option) (*zap.Logger, error) {
	config := zap.NewProductionConfig()
	config.DisableStacktrace = true
	config.EncoderConfig.TimeKey = "ts"
	config.EncoderConfig.EncodeTime = zapcore.RFC3339TimeEncoder
	return config.Build(opts...)
}
