package httpapi

import (
	"context"
	"encoding/json"
	"net/http"

	"github.com/gorilla/mux"
	"go.uber.org/zap"

	"code.vereign.com/gaiax/tsa/infohub/internal/tasks"
	"code.vereign.com/seal/backend/golib.git/errors"
)

type apiHandler struct {
	logger *zap.Logger
	tasks  TasksController
}

type TasksController interface {
	GetData(ctx context.Context, key string) (*tasks.GetDataResponse, error)
}

func New(tasks TasksController, logger *zap.Logger) http.Handler {
	handler := &apiHandler{
		logger: logger,
		tasks:  tasks,
	}

	router := mux.NewRouter()
	router.Path("/alive").HandlerFunc(handler.Alive).Methods("GET")
	router.Path("/ready").HandlerFunc(handler.Ready).Methods("GET")
	router.Path("/v1/info").HandlerFunc(handler.GetInfo).Methods("POST")

	return router
}

func (h *apiHandler) GetInfo(w http.ResponseWriter, r *http.Request) {
	var req GetInfoRequest

	if err := json.NewDecoder(r.Body).Decode(&req); err != nil {
		h.logger.Error("error decoding request", zap.Error(err))
		errors.JSON(w, errors.New(errors.BadRequest, "invalid request"))
		return
	}

	res, err := h.tasks.GetData(r.Context(), req.Data)
	if err != nil {
		h.logger.Error("error getting data from task controller", zap.Error(err))
		errors.JSON(w, errors.New(errors.Internal, "failed to get data"))
		return
	}

	response := &GetInfoResponse{
		ID:   res.ID,
		Data: res.Data,
	}

	if err = json.NewEncoder(w).Encode(response); err != nil {
		h.logger.Error("infohub get data: failed to send response", zap.Error(err))
	}
}

func (h *apiHandler) Alive(w http.ResponseWriter, r *http.Request) {
	h.encodeJSON(w, &struct {
		Status string `json:"status"`
	}{"up"})
}

func (h *apiHandler) Ready(w http.ResponseWriter, r *http.Request) {
	h.encodeJSON(w, &struct {
		Status string `json:"status"`
	}{"up"})
}

func (h *apiHandler) encodeJSON(w http.ResponseWriter, resp interface{}) {
	w.Header().Set("Content-Type", "application/json")
	_ = json.NewEncoder(w).Encode(resp)
}
